unit fMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, ColorGrd, StdCtrls, uEngine, ComCtrls;

type
  TfrmMain = class(TForm)
    timerGo: TTimer;
    panRight: TPanel;
    butLoad: TButton;
    butSave: TButton;
    butRandom: TButton;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    radBreed: TRadioButton;
    radWatch: TRadioButton;
    butAbout: TButton;
    tbMutationSpeed: TTrackBar;
    lblMutSpeed: TLabel;
    lblHint: TLabel;
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure timerGoTimer(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormCreate(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure butLoadClick(Sender: TObject);
    procedure butRandomClick(Sender: TObject);
    procedure butSaveClick(Sender: TObject);
    procedure radBreedClick(Sender: TObject);
    procedure FormCanResize(Sender: TObject; var NewWidth,
      NewHeight: Integer; var Resize: Boolean);
    procedure radWatchClick(Sender: TObject);
    procedure butAboutClick(Sender: TObject);
  private
    { Private declarations }
    FTerminated : Boolean;
    procedure ShowBreedingComponents;
  public
     FdX, FdY : Integer;
     FWorlds : array[0..1,0..1] of TWorld;
     procedure Restart;
  end;

var
  frmMain: TfrmMain;

implementation


uses
    uAbout;


{$R *.DFM}

procedure TfrmMain.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
    FTerminated := True;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
var
    X,Y : Integer;
begin
    SetPriorityClass( GetCurrentProcess, IDLE_PRIORITY_CLASS );
    ClientWidth := 4 * kMaxX + panRight.Width;
    ClientHeight := 4 * kMaxY;
    ShowBreedingComponents;
    for X := 0 to 1 do for Y := 0 to 1 do begin
        FWorlds[X,Y] := TWorld.Create;
    end;
end;

procedure TfrmMain.timerGoTimer(Sender: TObject);
var
    ci : Integer;
    R, X2,Y2,X1,Y1,X,Y,I : Integer;
    Animal : Array[0..kMaxAnimal-1] of record X,Y, Age : Integer end;
    avgInner, avgOuter : TCell;
    cred, cgreen, cblue : Integer;
    c1,c2 : TColor;
    pCurCell : ^TCell;
    dX,dY : Integer;
    rgb1,rgb2 : TRgb;
    C : TColor;
    fac : Double;
    W,H : Integer;
begin
    timerGo.Enabled := False;
    FTerminated := False;

    ci := 0;


    while not FTerminated do begin
        W := (ClientWidth - panRight.Width) div 2;
        H := clientHeight div 2;
        if radWatch.Checked then begin
            W := W * 2;
            H := H * 2;
        end;
        for X := 0 to 1 do for Y := 0 to 1 do begin
            if radWatch.Checked and (X+Y>0) then Continue;
            FWorlds[X,Y].Generation;
            FWorlds[X,Y].Render( Canvas, Bounds(X*W, Y*W,W,H) );
        end;

        Application.ProcessMessages;
    end;
end;

procedure TfrmMain.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
    W, H : Integer;
    X1,Y1 : Integer;
begin
    if radWatch.Checked then Exit;
    W := (ClientWidth - panRight.Width) div 2;
    H := clientHeight div 2;
    X := X div W;
    Y := Y div H;
    if (X<0) or (Y<0) or (X>1) or (Y>1) then Exit;
    for X1 := 0 to 1 do for Y1 := 0 to 1 do begin
        if (X1<>X) or (Y1<>Y) then
            FWorlds[X1,Y1].BadCopy(FWorlds[X,Y], tbMutationSpeed.Position);
        FWorlds[X1,Y1].FillRandom;
    end;
end;

procedure TfrmMain.Restart;
var
    X1,Y1 : Integer;
begin
    FTerminated := True;
    timerGo.Enabled := True;
    for X1 := 0 to 1 do for Y1 := 0 to 1 do begin
        FWorlds[X1,Y1].FillRandom;
    end;
end;


procedure TfrmMain.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
    if Key = VK_UP then Dec(FdY);
    if Key = VK_LEFT then Dec(FdX);
    if Key = VK_DOWN then Inc(FdY);
    if Key = VK_RIGHT then Inc(FdX);
end;

procedure TfrmMain.butRandomClick(Sender: TObject);
var
    X,Y : Integer;
begin
    for X := 0 to 1 do
        for Y := 0 to 1 do begin
            FWorlds[X,Y].RandomMatrix;
            FWorlds[X,Y].FillRandom;
        end;
    Restart;
end;

procedure TfrmMain.butLoadClick(Sender: TObject);
var
    FN : String;
    X,Y : Integer;
begin
    OpenDialog1.InitialDir := ExtractFileDir( ParamStr(0) );
    if not OpenDialog1.Execute then Exit;
    FN := OpenDialog1.FileName;
    FWorlds[0,0].LoadFromFile( FN );
    for X := 0 to 1 do for Y := 0 to 1 do begin
        if X+Y=0 then Continue;
        FWorlds[X,Y].BadCopy( FWorlds[0,0], tbMutationSpeed.Position );
    end;
    Restart;
end;

procedure TfrmMain.butSaveClick(Sender: TObject);
var
    FN : String;
begin
    SaveDialog1.InitialDir := ExtractFileDir( ParamStr(0) );
    if not SaveDialog1.Execute then Exit;
    FN := SaveDialog1.FileName;
    FWorlds[0,0].SaveToFile(FN);
end;

procedure TfrmMain.radBreedClick(Sender: TObject);
var
    X1,Y1 : Integer;
begin
    ShowBreedingComponents;
    for X1 := 0 to 1 do for Y1 := 0 to 1 do begin
        if (X1<>0) or (Y1<>0) then begin
            FWorlds[X1,Y1].BadCopy(FWorlds[0,0], tbMutationSpeed.Position);
            FWorlds[X1,Y1].CopyContents( FWorlds[0,0] );
        end;
    end;
end;

procedure TfrmMain.FormCanResize(Sender: TObject; var NewWidth,
  NewHeight: Integer; var Resize: Boolean);
var
    W1,H1, W2, H2 : Integer;
    MinF, F : Integer;
begin
    W1 := NewWidth - panRight.Width - Width + ClientWidth;
    H1 := NewHeight - Height + ClientHeight;

    MinF := (tbMutationSpeed.Top + tbMutationSpeed.Height + kMaxX-1) div kMaxY;

    F := ((W1 div kMaxX) + (H1 div kMaxY)) div 2;
    if F<MinF then F := MinF;

    NewWidth := F * kMaxX + panRight.Width + Width - ClientWidth;
    NewHeight := F * kMaxY  + Height - ClientHeight;
end;

procedure TfrmMain.radWatchClick(Sender: TObject);
begin
    ShowBreedingComponents;
end;

procedure TfrmMain.ShowBreedingComponents;
begin
    lblMutSpeed.Visible := radBreed.Checked;
    lblHint.Visible := radBreed.Checked;
    tbMutationSpeed.Visible := radBreed.Checked;
end;

procedure TfrmMain.butAboutClick(Sender: TObject);
begin
    frmAbout.ShowModal;
end;

initialization
    Randomize;
end.
