unit uEngine;

interface

uses
    Windows, Graphics;

const
    kMaxX = 64;
    kMaxY = kMaxX;
    kMaxAnimal = 70;
    kMultFac = 16;
    kShifter = 1 shl 16;


type
    TColor = (red,green,blue, yellow, magenta, cyan);
    TCell = array[TColor] of Integer;
    Trgb = Record R,G,B : Integer End;

    TMatrix = array[TColor, TColor] of Integer;

    TGrid = array[0..kMaxX-1, 0..kMaxY-1] of TCell;

    TWorld = class
      private
        FGrid : TGrid;
        FBack : TGrid;
        FMatrix : TMatrix;
        FBM : TBitmap;
        FInnerR, FScanSize : Integer;
      public
        constructor Create;
        destructor Destroy; override;
        procedure RandomMatrix;
        procedure FillRandom;
        procedure Generation;
        procedure BadCopy( From : TWorld; MutSpeed : Integer);
        procedure CopyContents( From : TWorld );
        procedure Render( ACanvas : TCanvas; Where : TRect );
        procedure LoadFromFile( FN : String );
        procedure SaveToFile( FN : String );
    end;

const
    kFull = 145;
    kHalf = 110;
    rgbMap : array[ TColor ] of TRgb = (
       ( R: kFull; G :     0; B :   0 ),
       ( R:     0; G : kFull; B :   0 ),
       ( R:     0; G :     0; B : kFull ),
       ( R: kHalf; G : kHalf; B :   0 ),
       ( R: kHalf; G :     0; B : kHalf ),
       ( R:     0; G : kHalf; B : kHalf )
    );


implementation

{ TWorld }
constructor TWorld.Create;
begin
    FBM := TBitmap.Create;
    FBM.Width := kMaxX;
    FBM.Height := kMaxY;
    FBM.HandleType := bmDIB;
    FBM.PixelFormat := pf24Bit;
    FillRandom;
    RandomMatrix;
end;

destructor TWorld.Destroy;
begin
    FBM.Free;
    inherited;
end;


procedure TWorld.BadCopy(From: TWorld; MutSpeed : Integer);
var
    C1,C2 : TColor;
begin
    FInnerR := From.FInnerR;
    FScanSize := From.FScanSize;
    for C1 := Low(TColor) to High(TColor) do begin
        for C2 := Low(TColor) to High(TColor) do begin
            FMatrix[C1,C2] :=  From.FMatrix[C1,C2] + Round( (Random(MutSpeed*2+1)-MutSpeed) * kShifter/10000);
        end;
    end;
end;

procedure TWorld.CopyContents(From: TWorld);
begin
    FGrid := From.FGrid;
    FBack := From.FBack;
end;

procedure TWorld.FillRandom;
var
    X,Y : Integer;
    C : TColor;
begin
    for X := 0 to kMaxX-1 do
        for Y := 0 to kMaxY-1 do
            for C := Low(TColor) to High(TColor) do
                FGrid[X,Y,c] := (Random(21)-10)*kMultFac;
end;

procedure TWorld.Generation;
var
    R, X2,Y2,X,Y : Integer;
    avgInner : TCell;
    c1,c2 : TColor;
    pCurCell : ^TCell;
    dX,dY : Integer;
    C : TColor;
begin
    FBack := FGrid;
    for X := 0 to kMaxX-1 do begin
        for Y := 0 to kMaxY-1 do begin
            for C := Low(TColor) to High(TColor) do begin
                avgInner[c] := 0;
            end;
            for dX := -FScanSize to +FScanSize do begin
                for dY:= -FScanSize to +FScanSize do begin
                    R := (abs(dX)+abs(dY));
                    if R=0 then Continue;
                    X2 := (dX + X + kMaxX) mod kMaxX;
                    Y2 := (dY + Y + kMaxY) mod kMaxY;
                    if R <= FInnerR then
                        for C := Low(TColor) to High(TColor) do Inc(avgInner[c],FBack[X2,Y2,C])
                end;
            end;
            pCurCell := @FGrid[X,Y];
            for C1 := Low(TColor) to High(TColor) do begin
                for C2 := Low(TColor) to High(TColor) do begin
                     if FMatrix[C1,C2]<>0 then
                         pCurCell^[C1] := pCurCell^[C1] + ( (avgInner[c2] * FMatrix[C1,C2]) div kShifter);
                end;
                if pCurCell^[C1]<-127*kMultFac then pCurCell^[C1]:=-127*kMultFac
                else if pCurCell^[C1]>127*kMultFac then pCurCell^[C1]:=127*kMultFac;
            end;
        end;
    end;

end;

procedure TWorld.LoadFromFile(FN: String);
var
    F : FILE OF TMatrix;
begin
    AssignFile( F, FN );
    Reset( F );
    Read( F, FMatrix );
    CloseFile( F);
end;

procedure TWorld.RandomMatrix;
var
    C1,C2 : TColor;
begin
    FillChar(FMatrix,SizeOf(FMatrix),#0);
        case Random(4) of
           0 : begin  FScanSize := 2;
                      FInnerR := 3;
               end;
           1 : begin  FScanSize := 1;
                      FInnerR := 2;
               end;
           2 : begin  FScanSize := 1;
                      FInnerR := 1;
               end;
           3 : begin  FScanSize := 3;
                      FInnerR := 5;
               end
        end;


    for C1 := Low(TColor) to High(TColor) do begin
        for C2 := Low(TColor) to High(TColor) do begin
            FMatrix[C1,C2] :=  Round( (Random(201)-100) * kShifter/10000);
        end;
    end;

end;

procedure TWorld.Render(ACanvas: TCanvas; Where : TRect);
var
    X1,Y1,X,Y : Integer;
    c1 : TColor;
    rgb1,rgb2 : TRgb;
    fac : Double;
begin
    // render the bitmap
    for X := 0 to kMaxX-1 do
        for Y := 0 to kMaxY-1 do begin
             X1 := (kMaxX + X) mod kMaxX;
             Y1 := (kMaxY + Y) mod kMaxY;
             FillChar( rgb1, SizeOf(rgb1), #0 );
             for C1 := Low(TColor) to High(TColor) do begin
                 rgb2 := rgbMap[C1];
                 fac := (FGrid[X,Y][C1]) / (kMultFac*256);
                 Inc(rgb1.r, Round( fac * rgb2.r ) );
                 Inc(rgb1.g, Round( fac * rgb2.g ) );
                 Inc(rgb1.b, Round( fac * rgb2.b ) );
             end;
             Inc(rgb1.r, 128 );
             Inc(rgb1.g, 128 );
             Inc(rgb1.b, 128 );
             if rgb1.r > 255 then rgb1.r := 255 else if rgb1.r <0 then rgb1.r := 0;
             if rgb1.g > 255 then rgb1.g := 255 else if rgb1.g <0 then rgb1.g := 0;
             if rgb1.b > 255 then rgb1.b := 255 else if rgb1.b <0 then rgb1.b := 0;
             FBM.Canvas.Pixels[X1,Y1] := Rgb( rgb1.r, rgb1.g, rgb1.b );
        end;

    StretchBlt( ACanvas.Handle, Where.Left, Where.Top, Where.Right - Where.Left, Where.Bottom - Where.Top,
                FBM.Canvas.Handle, 0,0, kMaxX, kMaxY,SRCCOPY );
end;

procedure TWorld.SaveToFile(FN: String);
var
    F : FILE OF TMatrix;
begin
    AssignFile( F, FN );
    Rewrite( F );
    Write( F, FMatrix );
    CloseFile( F);
end;

end.
